import os

if ("EMTRACKS_DDIR" not in os.environ) or ("EMTRACKS_PDIR" not in os.environ):
    print("ERROR! Please set $EMTRACKS_DDIR and $EMTRACKS_PDIR. Setting defaults (current directory)")
    emtracks_ddir = os.getcwd()+'/'
    emtracks_pdir = os.getcwd()+'/'
else:
    emtracks_ddir = os.environ["EMTRACKS_DDIR"]
    emtracks_pdir = os.environ["EMTRACKS_PDIR"]



import numpy as np

def get_B_df_distorted(df, v="0", **kwargs):
    if v == "0":
        return dist0(df, **kwargs)

def dist0(df, **kwargs):
    df_ = df.copy()

    if 'Bzf' in kwargs.keys():
        Bzf = kwargs['Bzf']
    else:
        Bzf = 0.
    if 'Bz0' in kwargs.keys():
        Bz0 = kwargs['Bz0']
    else:
        Bz0 = 100. # Gauss
    if 'z0' in kwargs.keys():
        z0 = kwargs['z0']
    else:
        z0 = 3.239 # m
    if 'zf' in kwargs.keys():
        zf = kwargs['zf']
    else:
        zf = 14.139 # m

    slope = (Bzf - Bz0) / (zf - z0)
    intercept = (Bzf - slope * zf)

    Bx, By, Bz = linear_gradient(df_[['X','Y','Z']].values.T, slope, intercept)
    # print(Bx.mean(),By.mean(),Bz.mean())
    Br = np.sqrt(Bx**2 + By**2)
    Bphi = -Bx*np.sin(df_.Phi.values)+By*np.cos(df_.Phi.values)
    df_.eval('Bx = Bx + @Bx', inplace=True)
    df_.eval('By = By + @By', inplace=True)
    df_.eval('Bz = Bz + @Bz', inplace=True)
    df_.eval('Br = Br + @Br', inplace=True)
    df_.eval('Bphi = Bphi + @Bphi', inplace=True)

    return df_

def linear_gradient(pos, slope, intercept):
    x, y, z = pos
    r = (x**2 + y**2)**(1/2)
    phi = np.arctan2(y,x)
    Bz = slope * z + intercept
    Br = - (r / 2) * slope #  Br = - (r / 2) * (d Bz / d Br)
    Bx = Br * np.cos(phi)
    By = Br * np.sin(phi)
    return np.array([Bx, By, Bz])



import matplotlib.pyplot as plt

# nicer plot formatting
def config_plots():
    #must run twice for some reason (glitch in Jupyter)
    for i in range(2):
        plt.rcParams['figure.figsize'] = [10, 8] # larger figures
        plt.rcParams['axes.grid'] = True         # turn grid lines on
        plt.rcParams['axes.axisbelow'] = True    # put grid below points
        plt.rcParams['grid.linestyle'] = '--'    # dashed grid
        plt.rcParams.update({'font.size': 12.0})   # increase plot font size

# config_plots()



import time
import math
import copy
import numpy as np
import pandas as pd
import dill as pkl
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from scipy.constants import c, elementary_charge
from scipy.integrate import solve_ivp
import scipy.optimize as optimize

import pypdt
from .helix import *
from .conversions import one_gev_c2_to_kg, one_kgm_s_to_mev_c, q_factor
from .plotting import config_plots
config_plots()


class trajectory_solver(object):
    def __init__(self, init_conds=None, bounds=None, particle_id=11, B_func=lambda p_vec: np.array([0.,0.,1.]), E_func = lambda p_vec: np.array([0., 0., 0.]), zevents=None):
        '''
        Default particle_id is for electron

        B_func should take in 3D numpy array of position, but defaults to a 1 Tesla field in the z direction.
        '''
        self.init_conds = init_conds
        self.bounds = bounds
        # dummy termination (no termination) if no bounds
        if bounds is None:
            self.terminate = lambda t, x: 1
        else:
            self.terminate = get_terminate(bounds)
        if particle_id < 0:
            sign = -1
        else:
            sign = +1
        particle_id = abs(particle_id)
        self.particle = pypdt.get(particle_id)
        self.particle.mass_kg = self.particle.mass * one_gev_c2_to_kg
        self.particle.charge_true = self.particle.charge*sign
        self.B_func = B_func
        self.E_func = E_func
        self.zevents = zevents

    @classmethod
    def from_pickle(cls, filename):
        return pkl.load(open(filename, "rb"))

    def to_pickle(self, filename):
        # delete B_func to avoid bloat
        # should include B_func meta data somewhere. FIX ME
        B_ = copy.deepcopy(self.B_func)
        del(self.B_func)
        # use dill package as "pkl" for picking lambda fcns
        pkl.dump(self, file=open(filename, 'wb'))
        # reset the B_func
        self.B_func = B_

    def z_event_func(self, t, y):
        # can define inside class since we don't need to set terminal=True
        z = y[2]
        if self.zevents is None:
            return 1
        # conds = [z == z_ for z_ in self.zevents]
        conds = np.isclose(z, self.zevents, atol=1e-2, rtol=1e-2)
        return int(not any(conds))

    def gamma(self, v_vec):
        '''
        Calculate gamma factor given a velocity (m / s),

        Args: v_vec is a 3 element np.array
        '''
        beta = v_vec / c
        return 1 / math.sqrt(1 - np.dot(beta, beta))

    def lorentz_accel(self, B_vec, E_vec, mom_vec, gamma):
        '''
        Calculate Lorentz acceleration
        '''
        a = self.particle.charge_true * elementary_charge * (E_vec*one_kgm_s_to_mev_c + 1. / (gamma*self.particle.mass_kg) * np.cross(mom_vec, B_vec))
        return a

    def lorentz_update(self, t, y):
        '''
        Function to have solve_ivp solve
        '''
        pos_vec = np.array(y[:3]) # meters
        mom_vec = np.array(y[3:]) # MeV / c
        v_vec = c * mom_vec / np.sqrt(np.dot(mom_vec, mom_vec) + (self.particle.mass*1000.)**2)
        B_vec = self.B_func(pos_vec) # B_vec based on position
        E_vec = self.E_func(pos_vec) # E_vec based on position
        gamma = self.gamma(v_vec) # gamma based on velocity
        f = [None] * 6 # [None, None, None, None, None, None]
        f[:3] = v_vec # y[3:] # position is updated by current velocity
        f[3:] = self.lorentz_accel(B_vec, E_vec, mom_vec, gamma)
        return f

    def solve_trajectory(self, verbose=False, method='DOP853', atol=1e-10, rtol=1e-10, dense=False):
        t_span = (self.init_conds.t0, self.init_conds.tf)
        t_eval = np.linspace(self.init_conds.t0, self.init_conds.tf, self.init_conds.N_t)
        p0 = self.init_conds.p0
        px0 = p0 * np.sin(self.init_conds.theta0) * np.cos(self.init_conds.phi0)
        py0 = p0 * np.sin(self.init_conds.theta0) * np.sin(self.init_conds.phi0)
        pz0 = p0 * np.cos(self.init_conds.theta0)
        y_init = [self.init_conds.x0, self.init_conds.y0, self.init_conds.z0, px0, py0, pz0]
        if self.zevents is None:
            events_list = [self.terminate]
        else:
            events_list = [self.terminate, self.z_event_func]
        if verbose:
            print(f"y_init: {y_init}")
        start_time = time.time()
        sol = solve_ivp(self.lorentz_update, t_span=t_span, y0 = y_init,\
                method=method, atol=atol, rtol=rtol, t_eval=t_eval,\
                events=events_list, dense_output=dense)
        t = sol.t
        t_e = sol.t_events
        x, y, z, px, py, pz = sol.y
        # x_e, y_e, z_e, px_e, py_e, pz_e = sol.y_events
        self.dataframe = pd.DataFrame({"t":t,"x":x,"y":y,"z":z,"px":px,"py":py,"pz":pz})
        # self.events_df = pd.DataFrame({"t":t_e,"x":x_e,"y":y_e,"z":z_e,"px":px_e,"py":py_e,"pz":pz_e})
        # calculate extra interesting variables
        # for df in [self.dataframe, self.events_df]:
        for df in [self.dataframe,]:
            df.eval("pT = (px**2 + py**2)**(1/2)", inplace=True)
            df.eval("p = (px**2 + py**2 + pz**2)**(1/2)", inplace=True)
            df.eval("E = (p**2 + (@self.particle.mass*1000.)**2)**(1/2)", inplace=True)
            df.eval("beta = p / E", inplace=True)
            df.eval("v = beta * @c", inplace=True)
            df.eval("vx = @c * px / E", inplace=True)
            df.eval("vy = @c * py / E", inplace=True)
            df.eval("vz = @c * pz / E", inplace=True)
            df.loc[:, "theta"] = np.arctan2(df.pT, df.pz)
            df.loc[:, "phi"] = np.arctan2(df.py, df.px)
        end_time = time.time()
        if verbose:
            print("Trajectory calculation complete!")
            print(f"Runtime: {(end_time - start_time):.4f} s")
        # return solution object for further use by user
        return sol

    def analyze_trajectory_LHelix(self, query="z >= 8.41 & z <= 11.66", step=100, stride=1):
        if query is None:
            df = self.dataframe.copy()
        else:
            df = self.dataframe.query(query)
        N_steps = int(len(df) // step)
        # pTs, pzs, ps, Es = [], [], [], []
        # pTs_nr, pzs_nr, ps_nr, Es_nr = [], [], [], []
        ps, ps_nr = [], []
        ps_guess, ps_guess_nr = [], []
        consistents = []
        df_fits = []
        df_fits_nr = []
        for i in range(N_steps):
            start = i * step
            stop = start + step
            df_ = df[start:stop:stride]
            track_data = df_[['t','x','y','z']].copy().values.T
            B_data = np.array([self.B_func(coord) for coord in track_data[1:].T]).T
            mom_fit, result, df_fit, params_fit, mom_guess, params_guess = fit_helix(track_data, B_data)
            mom_fit_nr, result_nr, df_fit_nr, params_fit_nr, mom_guess_nr, params_guess_nr = fit_helix(track_data, B_data, rot=False)
            if abs(mom_fit - mom_guess) / mom_guess < 1e-3:
                consistents.append(True)
            else:
                consistents.append(False)
            df_fits.append(df_fit)
            df_fits_nr.append(df_fit_nr)
            ps.append(mom_fit)
            ps_nr.append(mom_fit_nr)
            ps_guess.append(mom_guess)
            ps_guess_nr.append(mom_guess_nr)

        if len(df_fits) > 1:
            self.helix_reco_df = pd.concat(df_fits, ignore_index=True)
            self.no_rotation_helix_reco_df = pd.concat(df_fits_nr, ignore_index=True)
        else:
            self.helix_reco_df = df_fits
            self.no_rotation_helix_reco_df = df_fits_nr
        self.helix_reco_moms_df = pd.DataFrame({'p':ps, 'p_no_rotation':ps_nr, 'consistent':consistents})
        self.guess_helix_reco_moms_df = pd.DataFrame({'p':ps_guess, 'p_no_rotation':ps_guess_nr})
        if self.helix_reco_moms_df.consistent.sum() > 0:
            self.mom_LHelix = self.helix_reco_moms_df.query("consistent==True")['p'].mean()
            self.LHelix_success = True
        else:
            self.mom_LHelix = self.guess_helix_reco_moms_df['p'].mean()
            self.LHelix_success = False

    def analyze_trajectory(self, query=None, B=None, step=100, stride=10):
	# step = # rows to use for each arc segment
	# stride = points between each included reco point
	# N points in track reco = step // stride
        if B is None:
            B = self.B_func
        if query is None:
            df = self.dataframe.copy()
        else:
            df = self.dataframe.query(query)
        # print(df.head())

        N_steps = int(len(df) // step)
        charge_signs, masses, pTs, pzs, ps, Es, vs = [], [], [], [], [], [], []
        # ts_, ps_,
        # charge_signs, masses, pTs, pzs, ps, Es = [], [], [], [], [], []
        for i in range(N_steps):
            start = i * step
            stop = start + step
            df_ = df[start:stop:stride]
            ch, m, pT, pz, p, E, v = reco_arc(df_, B)
       	    charge_signs.append(ch)
            masses.append(m)
            pTs.append(pT)
            pzs.append(pz)
            ps.append(p)
            Es.append(E)
            vs.append(v)

        charge_signs = np.array(charge_signs)
        masses = np.array(masses)
        pTs = np.array(pTs)
        pzs = np.array(pzs)
        ps = np.array(ps)
        Es = np.array(Es)
        vs = np.array(vs)

        self.df_reco = pd.DataFrame({'p':ps, 'E':Es, 'm':masses, 'charge':charge_signs, 'pT':pTs, 'pz': pzs, 'v': vs})


    def plot3d(self, fig=None, ax=None, cmap="viridis"):
        if fig is None:
            fig = plt.figure(figsize=(12, 8))
            ax = fig.add_subplot(111, projection='3d')
        ax.plot(self.dataframe.x, self.dataframe.y, self.dataframe.z, 'k-', alpha=0.2, zorder=99)
        p = ax.scatter(self.dataframe.x, self.dataframe.y, self.dataframe.z, c=self.dataframe.t, cmap=cmap, s=2, alpha=1., zorder=101)
        cb = fig.colorbar(p)
        cb.set_label('t [s]', rotation=0.)
        ax.set_xlabel('\nX [m]', linespacing=3.0)
        ax.set_ylabel('\nY [m]', linespacing=3.0)
        ax.set_zlabel('\nZ [m]', linespacing=3.0)
        fig.tight_layout()
        def on_draw(event):
            reorder_camera_distance(ax, p)
        fig.canvas.mpl_connect('draw_event', on_draw)
        return fig, ax

    def plot2d(self):
        fig, axs = plt.subplots(3, 3)
        axs[0, 0].plot(self.dataframe.t, self.dataframe.x, 'bo-', markersize=2)
        axs[0, 0].set(xlabel="t [s]", ylabel="x [m]")
        axs[0, 1].plot(self.dataframe.t, self.dataframe.vx, 'bo-', markersize=2)
        axs[0, 1].set(xlabel="t [s]", ylabel="vx [m/s]")
        axs[0, 2].plot(self.dataframe.t, self.dataframe.px, 'bo-', markersize=2)
        axs[0, 2].set(xlabel="t [s]", ylabel="px [MeV/c]")
        axs[1, 0].plot(self.dataframe.t, self.dataframe.y, 'go-', markersize=2)
        axs[1, 0].set(xlabel="t [s]", ylabel="y [m]")
        axs[1, 1].plot(self.dataframe.t, self.dataframe.vy, 'go-', markersize=2)
        axs[1, 1].set(xlabel="t [s]", ylabel="vy [m/s]")
        axs[1, 2].plot(self.dataframe.t, self.dataframe.py, 'bo-', markersize=2)
        axs[1, 2].set(xlabel="t [s]", ylabel="py [MeV/c]")
        axs[2, 0].plot(self.dataframe.t, self.dataframe.z, 'ro-', markersize=2)
        axs[2, 0].set(xlabel="t [s]", ylabel="z [m]")
        axs[2, 1].plot(self.dataframe.t, self.dataframe.vz, 'ro-', markersize=2)
        axs[2, 1].set(xlabel="t [s]", ylabel="vz [m/s]")
        axs[2, 2].plot(self.dataframe.t, self.dataframe.pz, 'bo-', markersize=2)
        axs[2, 2].set(xlabel="t [s]", ylabel="pz [MeV/c]")
        fig.tight_layout()
        return fig, axs

# termination
def get_terminate(bounds):
    def terminate(t, y):
        pos = y[:3]
        conds = [pos[0]<bounds.xmin, pos[0]>bounds.xmax,
                 pos[1]<bounds.ymin, pos[1]>bounds.ymax,
                 pos[2]<bounds.zmin, pos[2]>bounds.zmax]
        return int(not any(conds))
    terminate.terminal = True
    return terminate

# PLOTTING ORDER FIX
def get_camera_position(ax):
    """returns the camera position for 3D axes in cartesian coordinates"""
    r = np.square(ax.xy_viewLim.max).sum()
    theta, phi = np.radians((90 - ax.elev, ax.azim))
    return np.array(sph2cart(r, theta, phi), ndmin=2).T

def sph2cart(r, theta, phi):
    """spherical to cartesian transformation."""
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)
    return x, y, z

def reorder_camera_distance(ax, patches):
    """
    Sort the patches (via their offsets) by decreasing distance from camera position
    so that the furthest gets drawn first.
    """
    # camera position in xyz
    camera = get_camera_position(ax)
    # distance of patches from camera
    d = np.square(np.subtract(camera, patches._offsets3d)).sum(0)
    o = d.argsort()[::-1]

    patches._offsets3d = tuple(np.array(patches._offsets3d)[:, o])
    patches._facecolor3d = patches._facecolor3d[o]
    patches._edgecolor3d = patches._edgecolor3d[o]
    # todo: similar for linestyles, linewidths, etc....

# def on_draw(event):
#     reorder_camera_distance()




import numpy as np
import pandas as pd
from scipy.constants import c
from .tools import InitConds
from .conversions import q_factor

def point_slope(x, xs, ys):
    slope = (ys[1]-ys[0])/(xs[1]-xs[0])
    y_result = slope*(x - xs[0]) + ys[0]
    return y_result

def interp_cole(df, z):
    if z > df.z.max() or z < df.z.min():
        raise ValueError('Invalid input z for interpolation. '+
                         f'Try again with {df.z.min():.3f} <= z <= {df.z.max():.3f}')
    i0 = df[df.z <= z].iloc[-1].name
    if i0 >= len(df)-1:
        i0 = len(df)-2
    interpolants = df.iloc[i0:i0+2]

    t_interps = interpolants.t.values
    x_interps = interpolants.x.values
    y_interps = interpolants.y.values
    z_interps = interpolants.z.values

    t_interp = point_slope(z, z_interps, t_interps)
    x_interp = point_slope(z, z_interps, x_interps)
    y_interp = point_slope(z, z_interps, y_interps)

    return t_interp, x_interp, y_interp


def get_ic_analytical(R, B, theta0, tf = 5e-8, N_t=5001):
    x0 = R
    y0 = 0.
    z0 = 0.
    t0 = 0.
    pT = q_factor * R * B
    pz = pT / np.tan(theta0)
    p0 = (pz**2 + pT**2)**(1/2)
    phi0 = np.pi/2
    ic = InitConds(t0=t0, tf=tf, N_t=N_t, x0=x0, y0=y0, z0=z0, p0=p0*1000., theta0=theta0, phi0=phi0)
    return ic

def interp_analytical(ic, m, z):
    # m == MeV/c
    R = ic.x0
    E = (ic.p0**2 + m**2)**(1/2)
    beta = ic.p0 / E
    v = beta * c
    vT = v * np.sin(ic.theta0)
    vz = v * np.cos(ic.theta0)
    period = 2*np.pi * R / vT
    const = 2*np.pi / period
    t_true = (z - ic.z0) / vz
    x_true = R * np.cos(const * t_true)
    y_true = R * np.sin(const * t_true)
    return t_true, x_true, y_true



#! /usr/bin/env python

import numpy as np
import pandas as pd
from scipy.interpolate import RegularGridInterpolator


def get_df_interp_func(filename=None, df=None, gauss=True, mm=False, scipy_interp=False, bounds=None, Blabels=['Bx','By','Bz']):
    '''
    This factory function will return an interpolating function for any field map. An input x,y,z will output the corresponding Bx,By,Bz or Br,Bphi,Bz. Will decide later if linear interpolation is good enough.

    Assumed file for input has length: meters, Bfield: Gauss
    '''
    # load dataframe if not passed in
    if df is None:
        if ".p" in filename:
            df = pd.read_pickle(filename)
        else:
            df = pd.read_csv(filename)
    else:
        df = df.copy()
    # labels
    Bx_l, By_l, Bz_l = Blabels
    if not gauss:
        df[Bx_l] = df[Bx_l] / 1e4
        df[By_l] = df[By_l] / 1e4
        df[Bz_l] = df[Bz_l] / 1e4
    if mm:
        df["X"] = df["X"] * 1e3
        df["Y"] = df["Y"] * 1e3
        df["Z"] = df["Z"] * 1e3

    xs = df.X.unique()
    ys = df.Y.unique()
    zs = df.Z.unique()

    if bounds is not None:
        xmin = xs[xs < bounds.xmin][-1]
        xmax = xs[xs > bounds.xmax][0]
        ymin = ys[ys < bounds.ymin][-1]
        ymax = ys[ys > bounds.ymax][0]
        zmin = zs[zs < bounds.zmin][-1]
        zmax = zs[zs > bounds.zmax][0]
        query_string = f"X>={xmin} & X<={xmax} & Y>={ymin} & Y<={ymax} & Z>={zmin} & Z<={zmax}"
        df = df.query(query_string)

        xs = df.X.unique()
        ys = df.Y.unique()
        zs = df.Z.unique()

    dx = xs[1]-xs[0]
    dy = ys[1]-ys[0]
    dz = zs[1]-zs[0]

    lx = len(xs)
    ly = len(ys)
    lz = len(zs)

    df_np = df[["X","Y","Z",Bx_l,By_l,Bz_l]].values

    x, y, z, bx, by, bz = df_np.T
    BX = bx.reshape(lx, ly, lz)
    BY = by.reshape(lx, ly, lz)
    BZ = bz.reshape(lx, ly, lz)

    interp_func_Bx = RegularGridInterpolator((xs, ys, zs), BX)
    interp_func_By = RegularGridInterpolator((xs, ys, zs), BY)
    interp_func_Bz = RegularGridInterpolator((xs, ys, zs), BZ)

    def scipy_interp_func(pts):
        Bx = interp_func_Bx(pts.T)
        By = interp_func_By(pts.T)
        Bz = interp_func_Bz(pts.T)
        return np.array([Bx, By, Bz])

    def get_cube(x, y, z):
        a_x, a_y, a_z = len(xs[xs <= x]) - 1, len(ys[ys <= y]) - 1, len(zs[zs <= z]) - 1
        corner_a = (ly * lz) * a_x + (lz) * a_y + a_z
        corner_b = corner_a + lz
        corner_c = corner_a + ly * lz
        corner_d = corner_a + ly * lz + lz
        index_list = [corner_a,corner_a+1,corner_b,corner_b+1,
        corner_c,corner_c+1,corner_d,corner_d+1]
        return df_np[index_list]

    def interp_single(xd,yd,zd,ff):
        c00 = ff[0,0,0]*(1 - xd) + ff[1,0,0] * xd
        c01 = ff[0,0,1]*(1 - xd) + ff[1,0,1] * xd
        c10 = ff[0,1,0]*(1 - xd) + ff[1,1,0] * xd
        c11 = ff[0,1,1]*(1 - xd) + ff[1,1,1] * xd

        c0 = c00 * (1 - yd) + c10 * yd
        c1 = c01 * (1 - yd) + c11 * yd
        return c0 * (1 - zd) + c1 * zd

    def interp(p_vec):
        cube = get_cube(*p_vec)

        xx = np.unique(cube[:,0])
        yy = np.unique(cube[:,1])
        zz = np.unique(cube[:,2])

        bxs_grid = cube[:,3].reshape((2,2,2))
        bys_grid = cube[:,4].reshape((2,2,2))
        bzs_grid = cube[:,5].reshape((2,2,2))

        xd = (p_vec[0]-xx[0])/(xx[1]-xx[0])
        yd = (p_vec[1]-yy[0])/(yy[1]-yy[0])
        zd = (p_vec[2]-zz[0])/(zz[1]-zz[0])

        bx = interp_single(xd,yd,zd, bxs_grid)
        by = interp_single(xd,yd,zd, bys_grid)
        bz = interp_single(xd,yd,zd, bzs_grid)

        return np.array([bx,by,bz])

    if scipy_interp:
        return scipy_interp_func
    else:
        return interp



# import numpy as np
import autograd.numpy as np
from autograd import grad
# from collections import namedtuple
from recordtype import recordtype

InitConds = recordtype('InitConds',['t0','tf','N_t','x0','y0','z0','p0','theta0','phi0'])
Bounds = recordtype('Bounds',['xmin', 'xmax', 'ymin', 'ymax', 'zmin', 'zmax'])

# a couple of canned examples
# default example
ic = InitConds(t0=0., tf=1e-8, N_t=1001, x0=0., y0=0., z0=0., p0=105., theta0=np.pi/6, phi0=0.)
bounds = Bounds(xmin=-0.95, xmax=0.95, ymin=-0.95, ymax=0.95, zmin=-1., zmax=1.)

# Mu2e examples
ic_Mu2e = InitConds(t0=0., tf=2e-7, N_t=20001,
                    x0=0.054094482, y0=0.03873037, z0=5.988900879,
                    p0=104.96, theta0=np.pi/3, phi0=0.)
ic_Mu2e_bounce = InitConds(t0=0., tf=2e-7, N_t=20001,
                           x0=0.054094482, y0=0.03873037, z0=5.988900879,
                           p0=104.96, theta0=np.pi-np.pi/3, phi0=0.)
bounds_Mu2e = Bounds(xmin=-0.95, xmax=0.95, ymin=-0.95, ymax=0.95, zmin=3.239, zmax=14.139)

# descent functions
def gradient_descent(g, alpha, max_its, w):
    # compute gradient module using autograd
    gradient = grad(g)
    # gradient descent loop
    weight_history = [w] # weight history container
    cost_history = [g(w)] # cost history container
    for k in range(max_its):
        # eval gradient
        grad_eval = gradient(w)
        # descent step
        w = w - alpha*grad_eval
        # record weight and cost
        weight_history.append(w)
        cost_history.append(g(w))

    return np.array(weight_history), np.array(cost_history)

def gradient_descent_dimalpha(g, alpha, max_its, w):
    # compute gradient module using autograd
    gradient = grad(g)
    # gradient descent loop
    weight_history = [w] # weight history container
    cost_history = [g(w)] # cost history container
    for k in range(max_its):
        # eval gradient
        grad_eval = gradient(w)
        # descent step
        w = w - alpha/(k+1)*grad_eval
        # record weight and cost
        weight_history.append(w)
        cost_history.append(g(w))

    return np.array(weight_history), np.array(cost_history)

def gradient_descent_mom_accel(g, alpha, beta, max_its, w):
    # compute gradient module using autograd
    gradient = grad(g)
    # gradient descent loop
    weight_history = [w] # weight history container
    cost_history = [g(w)] # cost history container
    d_vecs = []
    for k in range(max_its):
        # eval gradient
        grad_eval = gradient(w)
        # exponentially averaged descent direction
        if k == 0:
            d_vecs.append(-grad_eval)
        else:
            d_vecs.append(beta*d_vecs[-1] + (1-beta)*(-grad_eval))
        # descent step
        w = w + alpha*d_vecs[-1]
        # record weight and cost
        weight_history.append(w)
        cost_history.append(g(w))

    return np.array(weight_history), np.array(cost_history)



from numpy.polynomial.polynomial import Polynomial
from copy import deepcopy
import matplotlib.pyplot as plt
from .particle import trajectory_solver
from .tools import *
from .plotting import config_plots
config_plots()

def dist(sol_point, test_point):
    sol_point = sol_point[:3]
    return np.sqrt(np.sum((sol_point - test_point)**2))

def impact_param_point(sol_test, meas_point, alpha=1e-18, max_its=50, delta=0.001, plot_dists=False):
    # sol_test is solve_ivp solution object. Note must use dense=True to get a continuous solution
    # meas_point is a length 3 np.array denoting where our hypothesis track should pass through
    def g(t):
        return dist(sol_test.sol(t), meas_point)
    z_mask = np.array([False])
    i = 0
    while z_mask.sum() == 0:
        if i != 0:
            delta *= 2
        z_mask = (sol_test.y[2] > meas_point[2] - delta) & (sol_test.y[2] < meas_point[2] + delta)
        i += 1
    L = z_mask.sum()
    t0 = sol_test.t[z_mask][L // 2 - 1]
    wh, ch = gradient_descent_dimalpha(g, alpha, max_its, t0)
    tf = wh[ch.argmin()]
    if plot_dists:
        dists = np.array([g(t_) for t_ in sol_test.t])
        fig = plt.figure()
        plt.scatter(sol_test.t, dists, s=1, label='Data')
        plt.plot([t0, t0], [0., dists.max()], 'g-.', linewidth=1, label='Initial guess')
        plt.plot([tf, tf], [0., dists.max()], 'r--', linewidth=2, label='Optimized value')
        plt.xlabel('time [s]')
        plt.ylabel('distance between [m]')
        plt.legend()
    closest_point = sol_test.sol(tf)[:3]
    impact_param = dist(closest_point, meas_point)
    if plot_dists:
        return impact_param, closest_point, fig
    return impact_param, closest_point

def impact_params_in_tracker(sol_test, meas_solver, N_meas=40, tracker=True):
    if tracker:
        df = meas_solver.dataframe.query("z >= 8.41 & z <= 11.66").copy()
        df.reset_index(drop=True, inplace=True)
    else:
        df = meas_solver.dataframe

    stride = len(df) // N_meas
    df = df[::stride]
    impact_params = []
    closest_points = []
    for row in df.itertuples():
        ip, cp = impact_param_point(sol_test, np.array([row.x, row.y, row.z]))
        impact_params.append(ip)
        closest_points.append(cp)
    return np.array(impact_params), np.array(closest_points), df

def global_chi2_ptests(test_ps, test_B_func, meas_solver, sigma=1.):
    sum_of_impacts2 = []
    for p in test_ps:
        ic_ = deepcopy(meas_solver.init_conds)
        ic_.p0 = p
        test_solver = trajectory_solver(ic_, B_func=test_B_func, bounds=bounds_Mu2e)
        sol_test = test_solver.solve_trajectory(atol=1e-8, rtol=1e-8, dense=True)
        ips, cps, df_meas = impact_params_in_tracker(sol_test, meas_solver)
        sum_of_impacts2.append(np.sum(ips**2)/sigma**2)
    return np.array(sum_of_impacts2)

def momentum_analysis_global_chi2(meas_solver, analysis_B_func, B_name='Mau 13 (nominal)', sigma=1., delta=.2, N_hyp=7, plot_chi2=False):
    # use circle fit analysis mean to determine test_ps
    meas_solver.analyze_trajectory(step=25, stride=1, query="z >= 8.41 & z <= 11.66", B=analysis_B_func)
    circle_fit_mom = meas_solver.df_reco.p.mean()
    pmin = circle_fit_mom - delta
    pmax = circle_fit_mom + delta
    ps_test = np.linspace(pmin, pmax, N_hyp)
    # run particles for each hypothesis momentum
    sum_of_impacts2 = global_chi2_ptests(test_ps=ps_test, test_B_func=analysis_B_func,
                                         meas_solver=meas_solver, sigma=sigma)
    # fit quadratic to calculated chi2 values
    c, b, a = Polynomial.fit(ps_test, sum_of_impacts2, 2).convert().coef
    # use quadratic parameters to get momentum estimate
    chi2_fit_mom = - b / (2 * a)
    # plot (if selected)
    if plot_chi2:
        ptrue = meas_solver.init_conds.p0
        ps_quad = np.linspace(pmin, pmax, 50)
        y_quad = a * ps_quad**2 + b * ps_quad + c
        fig = plt.figure()
        plt.plot(ps_quad, y_quad, c='gray', label='Quadratic Fit')
        plt.scatter(ps_test, sum_of_impacts2, c='black', label='Calculated '+r"$\chi^2$", zorder=100)
        ymax = np.max(sum_of_impacts2)
        plt.plot([ptrue, ptrue], [0., ymax], 'r-', linewidth=2,
                 label=f'True Momentum ={ptrue:.3f} MeV/c', zorder=90)
        plt.plot([chi2_fit_mom, chi2_fit_mom], [0., ymax], 'g--',
                 label='Momentum '+r'$(\chi^2) =$'+f'{chi2_fit_mom:.3f} MeV/c', zorder=92)
        plt.plot([circle_fit_mom, circle_fit_mom], [0., ymax], 'b--',
                 label=f'Momentum (circle) = {circle_fit_mom:.3f} MeV/c', zorder=91)
        plt.xlabel(r"$p$"+" [MeV/c]")
        plt.ylabel(r"$\chi^2$")
        plt.title('Global '+r'$\chi^2$'+f' Momentum Fit: {B_name}')
        l = plt.legend(loc='upper right')
        l.set_zorder(110)
        return chi2_fit_mom, circle_fit_mom, fig
    # return results
    return chi2_fit_mom, circle_fit_mom



import math
import numbers
import numpy as np
import pandas as pd
import lmfit as lm
from scipy.spatial.transform import Rotation
import scipy.optimize as optimize

import pypdt
from .conversions import one_gev_c2_to_kg, one_kgm_s_to_mev_c, q_factor
from scipy.constants import c

cbar = c / 1.e9
cmmns = c / 1.e6
m_e = pypdt.get(11).mass*1000.
q_e = -1

# fitting routine
def fit_helix(track_data, B_data, m=m_e, q=q_e, rot=True):
    '''
    track_data is 4 by N np.array, B_data is 3 by N where each row is
    track_data: [t, x, y, z] with units [s, m, m, m]
    B_data: [Bx, By, Bz] with units [T, T, T]
    (to work out of the box with emtracks.particle)
    '''
    track_data = track_data.copy() # to not change track_data information
    track_data[0] = track_data[0]*1e9 # convert to ns
    track_data[1:] = track_data[1:]*1e3 # convert to mm
    # x0, y0 = track_data[1:3,0] # for phi0 estimate
    # translations:
    translate_vec = track_data[:,0].reshape(1, 4)
    track_data = track_data - np.repeat(translate_vec, track_data.shape[1], axis=0).T
    # Rotate so mean B vector is on z-axis
    B_mean = np.mean(B_data, axis=1)
    B = np.linalg.norm(B_mean)
    B_mean_T = np.linalg.norm(B_mean[:2]) # average transverse Bfield (original coords)
    # calculate rotation angles
    rot_theta = np.arctan2(B_mean_T, B_mean[2])
    rot_phi = np.arctan2(B_mean[1], B_mean[0])
    # create rotation function (+inverse) to align mean B with z axis
    rot_func = Rotation.from_euler('zy', np.array([-rot_phi, -rot_theta]))
    rot_inv_func = rot_func.inv()
    # rotate track data
    track_data_rot = track_data.copy()
    if rot:
        track_data_rot[1:] = rot_func.apply(track_data_rot[1:].T).T
    # estimate parameters
    # R, C_x, C_y
    cent, R_guess, Ri_fit, R_residual = reco_circle(track_data_rot[1], track_data_rot[2])
    C_x_guess, C_y_guess = cent
    # Lambda
    dists = (np.sum(np.square(track_data_rot[1:3]), axis=0))**(1/2)
    diffs_negative = np.diff(dists) < 0
    if diffs_negative.sum() == 0:
        endpoint = -1
    else:
        endpoint = np.argwhere(np.diff(dists) < 0).flatten()[0] - 1
    xyz_ends = track_data_rot[1:, [0, endpoint]]
    # print(xyz_ends)
    Lambda_guess = Lambda_est(R_guess, xyz_ends)
    # phi0
    # x0, y0 = track_data_rot[1:3,0] # WRONG-->[0,0,0]
    # x0p = -y0
    # y0p = x0
    # phi0_guess = np.arctan2(y0p, x0p)
    phi0_guess = abs(np.arctan2(C_x_guess, C_y_guess))
    # t0 should be 0 by construction (the translation)
    t0_guess = 0.
    # guess dict
    params_guess = {'R':R_guess, 'Lambda':Lambda_guess, 'C_x':C_x_guess, 'C_y':C_y_guess,
                    'phi0':phi0_guess, 't0':t0_guess}
    mom_guess = LHelix_get_momentum(params_guess['R'], params_guess['Lambda'], m, q, B)
    # construct model
    model = lm.Model(LHelix_P_pos, independent_vars=['t'])
    params = lm.Parameters()
    params.add('R', value=R_guess, min=R_guess-10, max=R_guess+10)
    params.add('Lambda', value=Lambda_guess, min=Lambda_guess-10, max=Lambda_guess+10)
    params.add('C_x', value=C_x_guess, min=C_x_guess-10, max=C_x_guess+10)
    params.add('C_y', value=C_y_guess, min=C_y_guess-10, max=C_y_guess+10)
    params.add('phi0', value=phi0_guess, min=phi0_guess-np.pi/10, max=phi0_guess+np.pi/10)
    # params.add('t0', value=t0_guess, min=t0_guess-1, max=t0_guess+1)#vary=False)
    # params.add('R', value=R_guess, )#min=R_guess-25, max=R_guess+25)
    # params.add('Lambda', value=Lambda_guess,) #min=Lambda_guess-25, max=Lambda_guess+25)
    # params.add('C_x', value=C_x_guess,) #min=C_x_guess-10, max=C_x_guess+10)
    # params.add('C_y', value=C_y_guess,) #min=C_y_guess-10, max=C_y_guess+10)
    # params.add('phi0', value=phi0_guess, min=0., max=2*np.pi)
    params.add('t0', value=t0_guess, vary=False)
    params.add('m', value=m, vary=False)
    params.add('q', value=q, vary=False)
    params.add('B', value=B, vary=False)

    result = model.fit(track_data_rot[1:], t=track_data_rot[0],params=params)
    params_fit = {key:val.value for key,val in result.params.items()}
    mom_fit = LHelix_get_momentum(params_fit['R'], params_fit['Lambda'], m, q, B)

    track_fit_xyz, track_fit_mom, track_fit_mom_vec = LHelix_P(track_data[0], **params_fit)
    # rotate track fit and momentum vec
    if rot:
        track_fit_xyz = rot_inv_func.apply(track_fit_xyz.T).T
        track_fit_mom_vec = rot_inv_func.apply(track_fit_mom_vec.T).T
    track_fit_xyz = track_fit_xyz + np.repeat(translate_vec[:,1:], track_data.shape[1], axis=0).T
    track_fit_xyz = 1e-3*track_fit_xyz
    df_fit = pd.DataFrame({'t':(track_data[0]+translate_vec[:,0])*1e-9, 'x':track_fit_xyz[0],
                           'y':track_fit_xyz[1], 'z':track_fit_xyz[2],
                           'px':track_fit_mom_vec[0], 'py':track_fit_mom_vec[1],
                           'pz':track_fit_mom_vec[2]})
    # return track_data_rot, R_guess, Lambda_guess, C_x_guess, C_y_guess, phi0_guess, t0_guess
    return mom_fit, result, df_fit, params_fit, mom_guess, params_guess
    # return mom_fit, result, params_fit, df_fit

# estimation helper functions
def Lambda_est(R_guess, xyz_ends):
    # R_guess from circle fit
    # xyz_ends 3 by 2 np.array where column 0 start point, column 1 end point
    delta_z = xyz_ends[2,1] - xyz_ends[2,0]
    chord_length = (np.sum(np.square(np.diff(xyz_ends[:2], axis=1))))**(1/2)
    # print(f'chord_length: {chord_length:.3f}, R_guess: {R_guess:.3f}')
    if 2*R_guess < chord_length:
        asin_ = np.pi/2
        # print('arcsin=pi/2')
    else:
        asin_ = np.arcsin(chord_length/(2*R_guess))
    theta_guess = np.arctan2(delta_z, 2*R_guess*asin_)
    return R_guess * np.tan(theta_guess)

# wrapper for position only
def LHelix_P_pos(**kwargs):
    return LHelix_P(**kwargs)[0]

# following helix parameterization used in KinKal (LHeliix, low momentum, arbitrary origin)
def LHelix_P(t, R, Lambda, C_x, C_y, phi0, t0, m, q, B):
    # [ns, mm, mm, mm, mm, rad, ns, MeV/c^2, integer*elementary_charge, Tesla
    dt = t - t0
    # find t format to get pz of correct shape/length
    if isinstance(t, numbers.Number):
        _pz_ones = 1.
    else:
        _pz_ones = np.ones(len(t))
    Q = -q*cbar*B
    mbar = m / Q
    ebar =  (R**2 + Lambda**2 + mbar**2)**(1/2) # name from KinKal

    Omega = math.copysign(1.,mbar) * cmmns / ebar
    x = C_x + R * np.sin(Omega*dt + phi0)
    y = C_y - R * np.cos(Omega*dt + phi0)
    z = Lambda * Omega * dt
    P_x = Q * R * np.cos(Omega*dt + phi0)
    P_y = Q * R * np.sin(Omega*dt + phi0)
    P_z = Q * Lambda * _pz_ones
    P_t = abs(Q) * ebar
    mom = P_t
    pos = np.array([x,y,z])
    mom_vec = np.array([P_x, P_y, P_z])
    return pos, mom, mom_vec

def LHelix_get_momentum(R, Lambda, m, q, B):
    Q = -q*cbar*B
    mbar = m / Q
    ebar =  (R**2 + Lambda**2 + mbar**2)**(1/2) # name from KinKal
    return abs(Q) * ebar

## CIRCLE FIT FUNCS
# circle fit
def calc_R(xc, yc, x, y):
    return np.sqrt((x-xc)**2 + (y-yc)**2)

def circ_alg_dist(center, x, y):
    Ri = calc_R(*center, x, y)
    return Ri - Ri.mean()

def reco_circle(x, y):
    x_m = np.mean(x)
    y_m = np.mean(y)
    center_est = x_m, y_m
    center_fit, ier = optimize.leastsq(circ_alg_dist, center_est, args=(x, y))
    Ri_fit = calc_R(*center_fit, x, y)
    R_fit = np.mean(Ri_fit)
    R_residual = np.sum((Ri_fit - R_fit)**2)
    return center_fit, R_fit, Ri_fit, R_residual

# full reco
def reco_arc(df, B_func):
    x, y, z, t, v, pT, pz, p, E = df[['x','y','z','t', 'v', 'pT', 'pz', 'p','E']].values.T
    # 1. reco circle
    center, R, Ri, res = reco_circle(x, y)
    # 2. Calculate theta
    thetai = np.arctan2(y-center[1], x-center[0])
    if not (thetai[0] <= 0. and thetai[-1] > 0.):
        thetai = (thetai + 2 * np.pi) % (2 * np.pi)
    arcangle = thetai[-1] - thetai[0]
    # 3. Calculate arc length
    arclength = R * arcangle
    # 4. Calculate z length
    G = np.array([np.ones_like(t), t]).T
    GtGinv = np.linalg.inv(G.T @ G)
    m = GtGinv @ G.T @ z
    # m[0] intercept, m[1] slope (aka speed in z direction, m / s)
    zlength = m[1] * (t[-1] - t[0])
    vz = m[1]
    # calculate vT
    vT = arclength / (t[-1] - t[0])
    # calculate v
    v = (vT**2 + vz**2)**(1/2)
    # calculate beta
    beta = v / c
    # 5. p, v, etc.
    tantheta = arclength / zlength
    gamma = (1 - beta**2)**(-1/2)
    Bxs, Bys, Bzs = np.array([B_func([xi,yi,zi]) for xi,yi,zi in zip(x,y,z)]).T
    Bs = (Bxs**2 + Bys**2 + Bzs**2)**(1/2)
    BTs = (Bxs**2 + Bys**2)**(1/2)
    # NEED TO FIND BEST B
    B = Bzs.mean()#Bs.mean() - BTs.mean()#Bzs.min()#Bzs.mean()
    pT = q_factor * B * R * 1000.
    pz = pT / tantheta
    p = (pT**2 + pz**2)**(1/2)
    mass = p * c / (gamma * v )
    E = (p**2 + mass**2)**(1/2)
    # charge
    charge_sign = - np.sign(m[1]) * np.sign(np.arctan2(arclength, zlength))
    return charge_sign, mass, pT, pz, p, E, v




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="emtracks",
    version="0.0.0",
    author="Cole Kampa",
    author_email="ckampa13@gmail.com",
    description="Particle in EM fields track calculator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FMS-Mu2e/particle_EM_tracks",
    packages=setuptools.find_packages(),
    install_requires=[
        "numpy", "scipy", "pandas", "matplotlib", "pypdt", "dill"
    ],
    python_requires='>=3.6',
    include_package_data=True,
    zip_safe=False,)



import math
import numpy as np
import pandas as pd
import scipy.optimize as optimization
from scipy import stats

import pypdt

# from digitization + normalization check
N_DIO_CD3 = 65 # 72
N_CE_CD3 = 10
p_range_CD3 = [102., 106.]

# e mass
m_e = 1000. * pypdt.get(11).mass

# DIO parameterization
a5, a6, a7, a8 = (8.6434e-17, 1.16874e-17, -1.87828e-19, 9.16327e-20)
E_mu = 105.194
E_endpoint = 104.973
m_al = 25133

def dio(mom_e):
    E_e = (mom_e**2 + m_e**2)**(1/2)
    # if E_e > E_endpoint:
    #     return 0.
    delta = E_mu - E_e - np.square(E_e) / (2 * m_al)
    f = a5 * delta**5 + a6 * delta**6 + a7 * delta**7 + a8 * delta**8
    if type(mom_e) is float:
        if E_e > E_endpoint:
            f = 0.
    else:
        f[E_e > E_endpoint] = 0.
    return f

# CE parameterization
CE_file = '/home/ckampa/data/root/ce_mom.csv'
mom = pd.read_csv(CE_file, names=['index', 'mom'])
mom = mom.mom
mom = mom[(mom > 98.) & (mom<110.)] # picks out right range for Crystal Ball fit
# run kde to use for fitter
positions = np.linspace(100., 104.9, 200)
kernel = stats.gaussian_kde(mom)
mom_kde = kernel(positions)

def crystal_ball(x, beta, m, loc, scale):
    return stats.crystalball.pdf(x, beta, m, loc, scale)

guess = np.array([1, 2, 104., 1.]) # from previous work (stats project)
x_dat = positions
params, cov = optimization.curve_fit(crystal_ball, x_dat, mom_kde, guess)
beta, m, loc, scale = params
y_max_ce = crystal_ball(loc, beta, m, loc, scale)

def ce(mom):
    return stats.crystalball.pdf(mom, beta, m, loc, scale)

# generator function
def generate_mu2e_event(N, interaction, low, high):
    # N: number of events to generate,
    # interaction: {"dio", "ce"}
    # efficiency: estimated sampling efficiency
    if interaction == "dio":
        ymax = dio(low)
        pdf_func = dio
        efficiency = 0.16
    elif interaction == "ce":
        ymax = y_max_ce
        pdf_func = ce
        efficiency = 0.6
    else:
        raise NotImplementedError('interaction is not in the list of supported interactions: ["ce", "dio"]')
    mom = np.array([])
    Ngen = math.ceil(N/efficiency) # account for sampling efficiency
    while len(mom) < N:
        xs = np.random.uniform(low=low, high=high, size=Ngen)
        ys = np.random.uniform(low=0, high=ymax, size=Ngen)
        passed = ys <= pdf_func(xs)
        mom = np.concatenate([mom, xs[passed]])
    return mom[:N]



from scipy.constants import c, elementary_charge

one_gev_c2_to_kg = 1.7826619e-27
one_kgm_s_to_mev_c = (1.7826619e-30 * c)**(-1)

q_factor = elementary_charge / one_gev_c2_to_kg / c

# p = 10 kg m / s
# p_MeV_c = p * one_kgm_s_to_mev_c
# p_kgm_s = p_MeV_c / one_kgm_s_to_mev_c
